;---------------------------------------
;
; Manber-Wu pattern matching
; Copyright (c) 1994 by Alessandro Scotti
;
;---------------------------------------
P386
JUMPS
LOCALS

MAXERRORS       EQU     8               ; Max error level for fuzzy search

PUBLIC  saExactSearch
PUBLIC  saFuzzySearch
PUBLIC  saSetFuzzy
PUBLIC  saSetPattern
PUBLIC  saGetCVT

;---------------------------------------
;
; Data segment
;
SE_DATA SEGMENT USE16 PARA PUBLIC 'DATA'
        ASSUME  ds:SE_DATA

; Data for saExactSearch()
CVector DD      256     DUP(?)     ; Characteristic vector
CMask   DD      ?                  ; Mask for a match
CMatch  DD      ?                  ; Current match status
PLength DW      ?                  ; Pattern length

; Data for saFuzzySearch()
FMatch  DD      MAXERRORS   DUP(?) ; Match table
OMatch  DD      MAXERRORS   DUP(?) ; Previous match table
ECount  DW      ?                  ; Max errors allowed (fuzzy search)

SE_DATA ENDS

;---------------------------------------
;
; Code segment
;
SE_TEXT SEGMENT USE16 PARA PUBLIC 'CODE'
        ASSUME  cs:SE_TEXT, es:NOTHING

;---------------------------------------
;
; Sets the pattern for a shift-and search.
;
; Input:
;       Pattern = pointer to pattern to search
;       PattLen = length of pattern
; Output:
;       zero if successful, otherwise a negative error code
;
saSetPattern    PROC PASCAL FAR
        ARG     Pattern:DWORD,  \
                PattLen:WORD
        USES    ds, si, es, di

        mov     ax, SEG SE_DATA
        mov     ds, ax
        mov     es, ax

; Check pattern length
        mov     dx, [PattLen]
        mov     ax, -1
        cmp     dx, 32                  ; Above maximum length?
        ja      @@Exit                  ; Yes, exit
        test    dx, dx                  ; Zero length?
        jz      @@Exit                  ; Yes, exit
        mov     [PLength], dx           ; Save pattern length

; Clear characteristic vector
        mov     di, OFFSET CVector
        movzx   esi, di
        xor     eax, eax
        mov     cx, 256
        dec     eax                     ; EAX = FFFFFFFFh
        cld
        rep     stosd

; Update characteristic vector with pattern information
        dec     al                      ; EAX = FFFFFFFEh
        les     di, [Pattern]
        xor     ebx, ebx
@@Loop:
        mov     bl, es:[di]             ; Get a character from pattern
        inc     di
        and     [esi+ebx*4], eax        ; Clear bit in current position
        rol     eax, 1                  ; Update bit mask
        dec     dx                      ; End of pattern?
        jnz     @@Loop                  ; No, get next character

; Save info and return
        ror     eax, 1
        not     eax
        mov     [CMask], eax            ; Save mask for match
        xor     eax, eax
        mov     [CMatch], eax           ; Clear match status
        xor     ax, ax                  ; Set return code

@@Exit:
        ret
saSetPattern    ENDP

;---------------------------------------
;
; Returns a pointer to the characteristic vector table.
;
; Input:
;       none
; Output:
;       a far pointer to the CVT
; Note:
;       case-insensitive matching, range of characters and wildcards
;       may be handled with manipulation of the CVT: no changes to
;       the search algorithm are required
;
saGetCVT        PROC PASCAL FAR
        mov     dx, SEG SE_DATA
        mov     ax, OFFSET CVector
        ret
saGetCVT        ENDP

;---------------------------------------
;
; Searches a buffer for an exact match of the current pattern.
;
; Input:
;       Buffer  = far pointer to buffer
;       BufLen  = buffer length in characters
; Output:
;       the index of the LAST character in a match, or -1 if no match found
; Note:
;       the meaning of "exact match" may be quite extended if the
;       CVT is properly modified before calling this function
;
saExactSearch   PROC PASCAL FAR
        ARG     Buffer:DWORD,   \
                BufLen:WORD
        USES    ds, si, es, di

        mov     ax, SEG SE_DATA
        mov     ds, ax

; Initialize registers
        mov     ax, -1                  ; No matches yet
        mov     cx, [BufLen]            ; Get buffer length
        jcxz    @@Exit                  ; Exit if buffer is empty
        cmp     cx, [PLength]           ; Buffer too small?
        jb      @@Exit                  ; Yes, exit
        les     di, [Buffer]            ; Get pointer to buffer
        mov     eax, [CMatch]           ; Get current match status
        mov     edx, [CMask]            ; Get mask for match
        mov     si, OFFSET CVector      ; Get offset of characteristic vector
        movzx   esi, si                 ; (avoids a 32-bit fixup)
        xor     ebx, ebx

; Search loop
@@Loop:
        mov     bl, es:[di]             ; Get character
        add     eax, eax                ; Shift current mask (EAX <<= 1)
        inc     di                      ; Increment buffer offset
        or      eax, [esi+ebx*4]        ; Mask with characteristic vector
        test    eax, edx                ; Does pattern match?
        jz      @@Done                  ; Yes, done
        dec     cx                      ; End of buffer?
        jnz     @@Loop                  ; No, continue

; Not found
        mov     [CMatch], eax           ; Update match status
        mov     ax, -1                  ; Assume pattern not found
        jmp     @@Exit                  ; End of buffer, exit

@@Done:
        mov     [CMatch], eax           ; Update match status
        mov     ax, di                  ; Get match offset
        sub     ax, WORD PTR [Buffer]
        dec     ax

@@Exit:
        cwd                             ; Convert return code to double word
        ret
saExactSearch   ENDP

;---------------------------------------
;
; Initializes approximate pattern matching.
;
; Input:
;       MaxErr  = max numbers of errors allowed (up to MAXERRORS-1)
; Output:
;       zero if successful, otherwise a negative error code
;
saSetFuzzy      PROC PASCAL FAR
        ARG     MaxErr:WORD
        USES    ds, es, di

        mov     ax, SEG SE_DATA
        mov     ds, ax
        mov     es, ax

        mov     ax, -1                  ; Assume an error
        mov     dx, [MaxErr]
        test    dx, dx                  ; Trying to do an exact search?
        jz      @@Exit                  ; Yes, exit
        cmp     dx, MAXERRORS           ; Is number of errors within range?
        jae     @@Exit                  ; No, exit

        shl     dx, 2                   ; Set error count
        mov     [ECount], dx

        xor     eax, eax                ; Init tables
        dec     eax                     ; EAX = FFFFFFFFh
        xor     edx, edx
        xor     bx, bx
        mov     cx, MAXERRORS
@@Loop:
        mov     FMatch[bx], eax
        mov     OMatch[bx], eax
        add     bx, 4
        shl     eax, 1
        dec     cx
        jnz     @@Loop

        xor     ax, ax                  ; Return success
@@Exit:
        ret
saSetFuzzy      ENDP

;---------------------------------------
;
; Searches a buffer for an approximate match of the current pattern.
;
; Input:
;       Buffer  = far pointer to buffer
;       BufLen  = buffer length in characters
; Output:
;       the index of the LAST character in a match, or -1 if not found
; Note:
;       because it's possible to have insertions and deletions, to find
;       where the matched pattern starts is not straightforward!
;
saFuzzySearch   PROC PASCAL FAR
        ARG     Buffer:DWORD,   \
                BufLen:WORD
        USES    ds, si, es, di

        mov     ax, SEG SE_DATA
        mov     ds, ax

; Initialize registers
        mov     ax, -1                  ; No matches yet
        mov     cx, [BufLen]            ; Get buffer length
        jcxz    @@Exit                  ; Exit if buffer is empty
        les     di, [Buffer]            ; Get pointer to buffer

; Loop thru buffer to find an approximated match
@@Loop:
        mov     eax, FMatch[0]          ; Get mask for exact match
        mov     bl, es:[di]             ; Get character
        xor     bh, bh                  ; Convert character to a word
        add     eax, eax                ; Shift
        shl     bx, 2                   ; Get offset into characteristic vector
        inc     di                      ; Bump character offset
        or      eax, CVector[bx]        ; Mask with characteristic vector
        mov     FMatch[0], eax          ; Save mask for exact match

; Update match mask for each error level
        mov     si, 4
@@FuzzyLoop:
        mov     eax, FMatch[si]         ; Get mask
        add     eax, eax                ; Handle insertion, deletion
        or      eax, CVector[bx]        ; and substitution (one more
        mov     edx, OMatch[si-4]       ; than the previous error level)
        and     eax, edx                ;
        and     edx, FMatch[si-4]       ;
        add     edx, edx                ;
        and     eax, edx                ;
        mov     FMatch[si], eax         ; Set mask
        add     si, 4                   ; Next error
        cmp     si, [ECount]            ; Max error level?
        jbe     @@FuzzyLoop             ; No, continue

; Update old masks (preserve EAX)
        sub     si, 4
@@FuzzyUpdateLoop:
        mov     edx, FMatch[si]
        mov     OMatch[si], edx
        sub     si, 4
        jge     @@FuzzyUpdateLoop

; We need only to test the mask for the maximum error level
        test    eax, [CMask]            ; Test max error level only!
        jz      @@Done                  ; Found a match, exit

; Loop thru all buffer characters
        dec     cx                      ; End of buffer?
        jnz     @@Loop                  ; No, continue

; Not found
        mov     ax, -1                  ; Assume pattern not found
        jmp     @@Exit                  ; End of buffer, exit

@@Done:
        mov     ax, di                  ; Get match offset
        sub     ax, WORD PTR [Buffer]
        dec     ax

@@Exit:
        cwd
        ret
saFuzzySearch   ENDP

SE_TEXT ENDS
END
